searchNodes=[{"doc":"Hammox is a library for rigorous unit testing using mocks, explicit behaviours and contract tests. See the README page for usage guide and examples. Most of the functions in this module come from Mox for backwards compatibility. As of v0.1.0, the only Hammox-specific functions are protect/2 and protect/3 .","ref":"Hammox.html","title":"Hammox","type":"module"},{"doc":"See Mox.allow/3 .","ref":"Hammox.html#allow/3","title":"Hammox.allow/3","type":"function"},{"doc":"See Mox.defmock/2 .","ref":"Hammox.html#defmock/2","title":"Hammox.defmock/2","type":"function"},{"doc":"See Mox.expect/4 .","ref":"Hammox.html#expect/4","title":"Hammox.expect/4","type":"function"},{"doc":"See protect/3 .","ref":"Hammox.html#protect/1","title":"Hammox.protect/1","type":"function"},{"doc":"See protect/3 .","ref":"Hammox.html#protect/2","title":"Hammox.protect/2","type":"function"},{"doc":"Decorates functions with Hammox checks based on given behaviour. Basic usage When passed an MFA tuple representing the function you'd like to protect, and a behaviour containing a callback for the function, it returns a new anonymous function that raises Hammox.TypeMatchError when called incorrectly or when it returns an incorrect value. Example: defmodule Calculator do @callback add ( integer ( ) , integer ( ) ) :: integer ( ) end defmodule TestCalculator do def add ( a , b ) , do : a + b end add_2 = Hammox . protect ( { TestCalculator , :add , 2 } , Calculator ) add_2 . ( 1.5 , 2.5 ) # throws Hammox.TypeMatchError Batch usage You can decorate all functions defined by a given behaviour by passing an implementation module and a behaviour module. Optionally, you can pass an explicit list of functions as the third argument. The returned map is useful as the return value for a test setup callback to set test context for all tests to use. Example: defmodule Calculator do @callback add ( integer ( ) , integer ( ) ) :: integer ( ) @callback add ( integer ( ) , integer ( ) , integer ( ) ) :: integer ( ) @callback add ( integer ( ) , integer ( ) , integer ( ) , integer ( ) ) :: integer ( ) @callback multiply ( integer ( ) , integer ( ) ) :: integer ( ) end defmodule TestCalculator do def add ( a , b ) , do : a + b def add ( a , b , c ) , do : a + b + c def add ( a , b , c , d ) , do : a + b + c + d def multiply ( a , b ) , do : a * b end %{ add_2 : add_2 , add_3 : add_3 , add_4 : add_4 multiply_2 : multiply_2 } = Hammox . protect ( TestCalculator , Calculator ) # optionally %{ add_2 : add_2 , add_3 : add_3 , multiply_2 : multiply_2 } = Hammox . protect ( TestCalculator , Calculator , add : [ 2 , 3 ] , multiply : 2 ) Batch usage for multiple behviours You can decorate all functions defined by any number of behaviours by passing an implementation module and a list of behaviour modules. The returned map is useful as the return value for a test setup callback to set test context for all tests to use. Example: defmodule Calculator do @callback add ( integer ( ) , integer ( ) ) :: integer ( ) @callback multiply ( integer ( ) , integer ( ) ) :: integer ( ) end defmodule AdditionalCalculator do @callback subtract ( integer ( ) , integer ( ) ) :: integer ( ) end defmodule TestCalculator do def add ( a , b ) , do : a + b def multiply ( a , b ) , do : a * b def subtract ( a , b ) , do : a - b end %{ add_2 : add_2 , multiply_2 : multiply_2 subtract_2 : subtract_2 } = Hammox . protect ( TestCalculator , [ Calculator , AdditionalCalculator ] ) Behaviour-implementation shortcuts Often, there exists one &quot;default&quot; implementation for a behaviour. A common practice is then to define both the callbacks and the implementations in one module. For these behaviour-implementation modules, Hammox provides shortucts that only require one module. Example: defmodule Calculator do @callback add ( integer ( ) , integer ( ) ) :: integer ( ) def add ( a , b ) , do : a + b end Hammox . protect ( { Calculator , :add , 2 } ) # is equivalent to Hammox . protect ( { Calculator , :add , 2 } , Calculator ) Hammox . protect ( Calculator , add : 2 ) # is equivalent to Hammox . protect ( Calculator , Calculator , add : 2 ) Hammox . protect ( Calculator ) # is equivalent to Hammox . protect ( Calculator , Calculator )","ref":"Hammox.html#protect/3","title":"Hammox.protect/3","type":"function"},{"doc":"See Mox.set_mox_from_context/1 .","ref":"Hammox.html#set_mox_from_context/1","title":"Hammox.set_mox_from_context/1","type":"function"},{"doc":"See Mox.set_mox_global/1 .","ref":"Hammox.html#set_mox_global/1","title":"Hammox.set_mox_global/1","type":"function"},{"doc":"See Mox.set_mox_private/1 .","ref":"Hammox.html#set_mox_private/1","title":"Hammox.set_mox_private/1","type":"function"},{"doc":"See Mox.stub/3 .","ref":"Hammox.html#stub/3","title":"Hammox.stub/3","type":"function"},{"doc":"See Mox.stub_with/2 .","ref":"Hammox.html#stub_with/2","title":"Hammox.stub_with/2","type":"function"},{"doc":"See Mox.verify!/0 .","ref":"Hammox.html#verify!/0","title":"Hammox.verify!/0","type":"function"},{"doc":"See Mox.verify!/1 .","ref":"Hammox.html#verify!/1","title":"Hammox.verify!/1","type":"function"},{"doc":"See Mox.verify_on_exit!/1 .","ref":"Hammox.html#verify_on_exit!/1","title":"Hammox.verify_on_exit!/1","type":"function"},{"doc":"","ref":"Hammox.html#t:function_arity_pair/0","title":"Hammox.function_arity_pair/0","type":"type"},{"doc":"A use able module simplifying protecting functions with Hammox. The explicit way is to use Hammox.protect/3 and friends to generate protected versions of functions as anonymous functions. In tests, the most convenient way is to generate them once in a setup hook and then resolve them from test context. However, this can get quite verbose. If you're willing to trade explicitness for some macro magic, doing use Hammox.Protect in your test module will define functions from the module you want to protect in it. The effect is similar to import ing the module you're testing, but with added benefit of the functions being protected. use Hammox.Protect supports these options: :module (required) — the module you'd like to protect (usually the one you're testing in the test module). Equivalent to the first parameter of Hammox.protect/3 in batch usage. :behaviour — the behaviour module you'd like to protect the implementation module with. Can be skipped if :module and :behaviour are the same module. Equivalent to the second parameter of Hammox.protect/3 in batch usage. :funs — An optional explicit list of functions you'd like to protect. Equivalent to the third parameter of Hammox.protect/3 in batch usage. Additionally multiple behaviour and funs options can be provided for modules that implement multiple behaviours note: the funs options are optional but specific to the behaviour that precedes them use Hammox.Protect , module : Hammox.Test.MultiBehaviourImplementation , behaviour : Hammox.Test.SmallBehaviour , # the `funs` opt below effects the funs protected from `SmallBehaviour` funs : [ foo : 0 , other_foo : 1 ] , behaviour : Hammox.Test.AdditionalBehaviour # with no `funs` pt provided after `AdditionalBehaviour`, all callbacks # will be protected ` ` ` `","ref":"Hammox.Protect.html","title":"Hammox.Protect","type":"module"},{"doc":"An error thrown when Hammox detects that values in a function call don't match types defined in typespecs.","ref":"Hammox.TypeMatchError.html","title":"Hammox.TypeMatchError","type":"exception"},{"doc":"Apache License Version 2.0, January 2004 http://www.apache.org/licenses/ TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION 1. Definitions. &quot;License&quot; shall mean the terms and conditions for use, reproduction, and distribution as defined by Sections 1 through 9 of this document. &quot;Licensor&quot; shall mean the copyright owner or entity authorized by the copyright owner that is granting the License. &quot;Legal Entity&quot; shall mean the union of the acting entity and all other entities that control, are controlled by, or are under common control with that entity. For the purposes of this definition, &quot;control&quot; means (i) the power, direct or indirect, to cause the direction or management of such entity, whether by contract or otherwise, or (ii) ownership of fifty percent (50%) or more of the outstanding shares, or (iii) beneficial ownership of such entity. &quot;You&quot; (or &quot;Your&quot;) shall mean an individual or Legal Entity exercising permissions granted by this License. &quot;Source&quot; form shall mean the preferred form for making modifications, including but not limited to software source code, documentation source, and configuration files. &quot;Object&quot; form shall mean any form resulting from mechanical transformation or translation of a Source form, including but not limited to compiled object code, generated documentation, and conversions to other media types. &quot;Work&quot; shall mean the work of authorship, whether in Source or Object form, made available under the License, as indicated by a copyright notice that is included in or attached to the work (an example is provided in the Appendix below). &quot;Derivative Works&quot; shall mean any work, whether in Source or Object form, that is based on (or derived from) the Work and for which the editorial revisions, annotations, elaborations, or other modifications represent, as a whole, an original work of authorship. For the purposes of this License, Derivative Works shall not include works that remain separable from, or merely link (or bind by name) to the interfaces of, the Work and Derivative Works thereof. &quot;Contribution&quot; shall mean any work of authorship, including the original version of the Work and any modifications or additions to that Work or Derivative Works thereof, that is intentionally submitted to Licensor for inclusion in the Work by the copyright owner or by an individual or Legal Entity authorized to submit on behalf of the copyright owner. For the purposes of this definition, &quot;submitted&quot; means any form of electronic, verbal, or written communication sent to the Licensor or its representatives, including but not limited to communication on electronic mailing lists, source code control systems, and issue tracking systems that are managed by, or on behalf of, the Licensor for the purpose of discussing and improving the Work, but excluding communication that is conspicuously marked or otherwise designated in writing by the copyright owner as &quot;Not a Contribution.&quot; &quot;Contributor&quot; shall mean Licensor and any individual or Legal Entity on behalf of whom a Contribution has been received by Licensor and subsequently incorporated within the Work. 2. Grant of Copyright License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable copyright license to reproduce, prepare Derivative Works of, publicly display, publicly perform, sublicense, and distribute the Work and such Derivative Works in Source or Object form. 3. Grant of Patent License. Subject to the terms and conditions of this License, each Contributor hereby grants to You a perpetual, worldwide, non-exclusive, no-charge, royalty-free, irrevocable (except as stated in this section) patent license to make, have made, use, offer to sell, sell, import, and otherwise transfer the Work, where such license applies only to those patent claims licensable by such Contributor that are necessarily infringed by their Contribution(s) alone or by combination of their Contribution(s) with the Work to which such Contribution(s) was submitted. If You institute patent litigation against any entity (including a cross-claim or counterclaim in a lawsuit) alleging that the Work or a Contribution incorporated within the Work constitutes direct or contributory patent infringement, then any patent licenses granted to You under this License for that Work shall terminate as of the date such litigation is filed. 4. Redistribution. You may reproduce and distribute copies of the Work or Derivative Works thereof in any medium, with or without modifications, and in Source or Object form, provided that You meet the following conditions: (a) You must give any other recipients of the Work or Derivative Works a copy of this License; and (b) You must cause any modified files to carry prominent notices stating that You changed the files; and (c) You must retain, in the Source form of any Derivative Works that You distribute, all copyright, patent, trademark, and attribution notices from the Source form of the Work, excluding those notices that do not pertain to any part of the Derivative Works; and (d) If the Work includes a &quot;NOTICE&quot; text file as part of its distribution, then any Derivative Works that You distribute must include a readable copy of the attribution notices contained within such NOTICE file, excluding those notices that do not pertain to any part of the Derivative Works, in at least one of the following places: within a NOTICE text file distributed as part of the Derivative Works; within the Source form or documentation, if provided along with the Derivative Works; or, within a display generated by the Derivative Works, if and wherever such third-party notices normally appear. The contents of the NOTICE file are for informational purposes only and do not modify the License. You may add Your own attribution notices within Derivative Works that You distribute, alongside or as an addendum to the NOTICE text from the Work, provided that such additional attribution notices cannot be construed as modifying the License. You may add Your own copyright statement to Your modifications and may provide additional or different license terms and conditions for use, reproduction, or distribution of Your modifications, or for any such Derivative Works as a whole, provided Your use, reproduction, and distribution of the Work otherwise complies with the conditions stated in this License. 5. Submission of Contributions. Unless You explicitly state otherwise, any Contribution intentionally submitted for inclusion in the Work by You to the Licensor shall be under the terms and conditions of this License, without any additional terms or conditions. Notwithstanding the above, nothing herein shall supersede or modify the terms of any separate license agreement you may have executed with Licensor regarding such Contributions. 6. Trademarks. This License does not grant permission to use the trade names, trademarks, service marks, or product names of the Licensor, except as required for reasonable and customary use in describing the origin of the Work and reproducing the content of the NOTICE file. 7. Disclaimer of Warranty. Unless required by applicable law or agreed to in writing, Licensor provides the Work (and each Contributor provides its Contributions) on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied, including, without limitation, any warranties or conditions of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A PARTICULAR PURPOSE. You are solely responsible for determining the appropriateness of using or redistributing the Work and assume any risks associated with Your exercise of permissions under this License. 8. Limitation of Liability. In no event and under no legal theory, whether in tort (including negligence), contract, or otherwise, unless required by applicable law (such as deliberate and grossly negligent acts) or agreed to in writing, shall any Contributor be liable to You for damages, including any direct, indirect, special, incidental, or consequential damages of any character arising as a result of this License or out of the use or inability to use the Work (including but not limited to damages for loss of goodwill, work stoppage, computer failure or malfunction, or any and all other commercial damages or losses), even if such Contributor has been advised of the possibility of such damages. 9. Accepting Warranty or Additional Liability. While redistributing the Work or Derivative Works thereof, You may choose to offer, and charge a fee for, acceptance of support, warranty, indemnity, or other liability obligations and/or rights consistent with this License. However, in accepting such obligations, You may act only on Your own behalf and on Your sole responsibility, not on behalf of any other Contributor, and only if You agree to indemnify, defend, and hold each Contributor harmless for any liability incurred by, or claims asserted against, such Contributor by reason of your accepting any such warranty or additional liability. END OF TERMS AND CONDITIONS APPENDIX: How to apply the Apache License to your work. To apply the Apache License to your work, attach the following boilerplate notice, with the fields enclosed by brackets &quot;[]&quot; replaced with your own identifying information. (Don't include the brackets!) The text should be enclosed in the appropriate comment syntax for the file format. We also recommend that a file or class name and description of purpose be included on the same &quot;printed page&quot; as the copyright notice for easier identification within third-party archives. Copyright 2019 Michał Szewczak Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"license.html","title":"License","type":"extras"},{"doc":"Hammox is a library for rigorous unit testing using mocks, explicit behaviours and contract tests. You can use it to ensure both your mocks and implementations fulfill the same contract. It takes the excellent Mox library and pushes its philosophy to its limits, providing automatic contract tests based on behaviour typespecs while maintaining full compatibility with code already using Mox. Hammox aims to catch as many contract bugs as possible while providing useful deep stacktraces so they can be easily tracked down and fixed.","ref":"readme.html","title":"Overview","type":"extras"},{"doc":"If you are currently using Mox , delete it from your list of dependencies in mix.exs . Then add :hammox : def deps do [ { :hammox , &quot;~&gt; 0.6&quot; , only : :test } ] end","ref":"readme.html#installation","title":"Overview - Installation","type":"extras"},{"doc":"Read &quot;Mocks and explicit contracts&quot; by José Valim. Then proceed to the Mox documentation . Once you are comfortable with Mox, switch to using Hammox.","ref":"readme.html#starting-from-scratch","title":"Overview - Starting from scratch","type":"extras"},{"doc":"Replace all occurrences of Mox with Hammox . Nothing more is required; all your mock calls in test are now ensured to conform to the behaviour typespec.","ref":"readme.html#migrating-from-mox","title":"Overview - Migrating from Mox","type":"extras"},{"doc":"Typical mock setup Let's say we have a database which can get us user data. We have a module, RealDatabase (not shown), which implements the following behaviour: defmodule Database do @callback get_users ( ) :: [ binary ( ) ] end We use this client in a Stats module which can aggregate data about users: defmodule Stats do def count_users ( database \\\\ RealDatabase ) do length ( database . get_users ( ) ) end end And we create a unit test for it: defmodule StatsTest do use ExUnit.Case , async : true test &quot;count_users/0 returns correct user count&quot; do assert 2 == Stats . count_users ( ) end end For this test to work, we would have to start a real instance of the database and provision it with two users. This is of course unnecessary brittleness — in a unit test, we only want to test that our Stats code provides correct results given specific inputs. To simplify, we will create a mocked Database using Mox and use it in the test: defmodule StatsTest do use ExUnit.Case , async : true import Mox test &quot;count_users/0 returns correct user count&quot; do defmock ( DatabaseMock , for : Database ) expect ( DatabaseMock , :get_users , fn -&gt; [ &quot;joe&quot; , &quot;jim&quot; ] end ) assert 2 == Stats . count_users ( DatabaseMock ) end end The test now passes as expected. The contract breaks Imagine that some time later we want to add error flagging for our database client. We change RealDatabase and the corresponding behaviour, Database , to return an ok/error tuple instead of a raw value: defmodule Database do @callback get_users ( ) :: { :ok , [ binary ( ) ] } | { :error , term ( ) } end However, The Stats.count_users/0 test will still pass , even though the function will break when the real database client is used! This is because the mock is now invalid — it no longer implements the given behaviour, and therefore breaks the contract. Even though Mox is supposed to create mocks following explicit contracts, it does not take typespecs into account. This is where Hammox comes in. Simply replace all occurrences of Mox with Hammox (for example, import Mox becomes import Hammox , etc) and you will now get this when trying to run the test: ** (Hammox.TypeMatchError) Returned value [&quot;joe&quot;, &quot;jim&quot;] does not match type {:ok, [binary()]} | {:error, term()}. Now the consistency between the mock and its behaviour is enforced. Completing the triangle Hammox automatically checks mocks with behaviours, but what about the real implementations? The real goal is to keep all units implementing a given behaviour in sync. You can decorate any function with Hammox checks by using Hammox.protect/2 . It will return an anonymous function which you can use in place of the original module function. An example test: defmodule RealDatabaseTest do use ExUnit.Case , async : true test &quot;get_users/0 returns list of users&quot; do get_users_0 = Hammox . protect ( { RealDatabase , :get_users , 0 } , Database ) assert { :ok , [ &quot;jim&quot; , &quot;joe&quot; ] } == get_users_0 . ( ) end end It's a good idea to put setup logic like this in a setup_all hook and then access the protected functions using the test context: defmodule RealDatabaseTest do use ExUnit.Case , async : true setup_all do %{ get_users_0 : Hammox . protect ( { RealDatabase , :get_users , 0 } , Database ) } end test &quot;get_users/0 returns list of users&quot; , %{ get_users_0 : get_users_0 } do assert { :ok , [ &quot;jim&quot; , &quot;joe&quot; ] } == get_users_0 . ( ) end end Hammox also provides a setup_all friendly version of Hammox.protect which leverages this pattern. Simply pass both the implementation module and the behaviour module and you will get a map of all callbacks defined by the behaviour as decorated implementation functions. defmodule RealDatabaseTest do use ExUnit.Case , async : true setup_all do Hammox . protect ( RealDatabase , Database ) end test &quot;get_users/0 returns list of users&quot; , %{ get_users_0 : get_users_0 } do assert { :ok , [ &quot;jim&quot; , &quot;joe&quot; ] } == get_users_0 . ( ) end end Alternatively, if you're up for trading explicitness for some macro magic, you can use use Hammox.Protect to locally define protected versions of functions you're testing, as if you import ed the module: defmodule RealDatabaseTest do use ExUnit.Case , async : true use Hammox.Protect , module : RealDatabase , behaviour : Database test &quot;get_users/0 returns list of users&quot; do assert { :ok , [ &quot;jim&quot; , &quot;joe&quot; ] } == get_users ( ) end end","ref":"readme.html#example","title":"Overview - Example","type":"extras"},{"doc":"Dialyzer cannot detect Mox style mocks not conforming to typespec. The main aim of Hammox is to enforce consistency between behaviours, mocks and implementations. This is best achieved when both mocks and implementations are subjected to the exact same checks. Dialyzer is a static analysis tool; Hammox is a dynamic contract test provider. They operate differently and one can catch some bugs when the other doesn't. While it is true that Hammox would be redundant given a strong, strict, TypeScript-like type system for Elixir, Dialyzer is far from providing that sort of coverage.","ref":"readme.html#why-use-hammox-for-my-application-code-when-i-have-dialyzer","title":"Overview - Why use Hammox for my application code when I have Dialyzer?","type":"extras"},{"doc":"A t() type defined on a protocol is taken by Hammox to mean &quot;a struct implementing the given protocol&quot;. Therefore, trying to pass :atom for an Enumerable.t() will produce an error, even though the type is defined as term() : ** (Hammox.TypeMatchError) Returned value :atom does not match type Enumerable.t(). Value :atom does not implement the Enumerable protocol.","ref":"readme.html#protocol-types","title":"Overview - Protocol types","type":"extras"},{"doc":"Hammox also includes Mox as a dependency. This means that if you would like to disable Hammox protection for a specific mock, you can simply use vanilla Mox for that specific instance. They will interoperate without problems.","ref":"readme.html#disable-protection-for-specific-mocks","title":"Overview - Disable protection for specific mocks","type":"extras"},{"doc":"For anonymous function types in typespecs, only the arity is checked. Parameter types and return types are not checked.","ref":"readme.html#limitations","title":"Overview - Limitations","type":"extras"},{"doc":"Copyright 2019 Michał Szewczak Licensed under the Apache License, Version 2.0 (the &quot;License&quot;); you may not use this file except in compliance with the License. You may obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0 Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an &quot;AS IS&quot; BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.","ref":"readme.html#license","title":"Overview - License","type":"extras"}]